"use strict";
var isArray = require('lodash.isarray');
var isNull = require('lodash.isnull');
var isString = require('lodash.isstring');
var has = require('lodash.has');
var parser_1 = require('../parser');
var storeUtils_1 = require('./storeUtils');
function diffQueryAgainstStore(_a) {
    var store = _a.store, query = _a.query, variables = _a.variables, dataIdFromObject = _a.dataIdFromObject;
    var queryDef = parser_1.parseQuery(query);
    return diffSelectionSetAgainstStore({
        store: store,
        rootId: 'ROOT_QUERY',
        selectionSet: queryDef.selectionSet,
        throwOnMissingField: false,
        variables: variables,
        dataIdFromObject: dataIdFromObject,
    });
}
exports.diffQueryAgainstStore = diffQueryAgainstStore;
function diffFragmentAgainstStore(_a) {
    var store = _a.store, fragment = _a.fragment, rootId = _a.rootId, variables = _a.variables, dataIdFromObject = _a.dataIdFromObject;
    var fragmentDef = parser_1.parseFragment(fragment);
    return diffSelectionSetAgainstStore({
        store: store,
        rootId: rootId,
        selectionSet: fragmentDef.selectionSet,
        throwOnMissingField: false,
        variables: variables,
        dataIdFromObject: dataIdFromObject,
    });
}
exports.diffFragmentAgainstStore = diffFragmentAgainstStore;
function diffSelectionSetAgainstStore(_a) {
    var selectionSet = _a.selectionSet, store = _a.store, rootId = _a.rootId, _b = _a.throwOnMissingField, throwOnMissingField = _b === void 0 ? false : _b, variables = _a.variables, dataIdFromObject = _a.dataIdFromObject;
    if (selectionSet.kind !== 'SelectionSet') {
        throw new Error('Must be a selection set.');
    }
    var result = {};
    var missingSelectionSets = [];
    var missingFields = [];
    var storeObj = store[rootId] || {};
    selectionSet.selections.forEach(function (selection) {
        if (selection.kind !== 'Field') {
            throw new Error('Only fields supported so far, not fragments.');
        }
        var field = selection;
        var storeFieldKey = storeUtils_1.storeKeyNameFromField(field, variables);
        var resultFieldKey = storeUtils_1.resultKeyNameFromField(field);
        var missingFieldPushed = false;
        function pushMissingField(missingField) {
            if (!missingFieldPushed) {
                missingFields.push(missingField);
                missingFieldPushed = true;
            }
        }
        if (!has(storeObj, storeFieldKey)) {
            if (throwOnMissingField) {
                throw new Error("Can't find field " + storeFieldKey + " on object " + storeObj + ".");
            }
            missingFields.push(field);
            return;
        }
        var storeValue = storeObj[storeFieldKey];
        if (!field.selectionSet) {
            result[resultFieldKey] = storeValue;
            return;
        }
        if (isNull(storeValue)) {
            result[resultFieldKey] = null;
            return;
        }
        if (isArray(storeValue)) {
            result[resultFieldKey] = storeValue.map(function (id) {
                if (isNull(id)) {
                    return null;
                }
                var itemDiffResult = diffSelectionSetAgainstStore({
                    store: store,
                    throwOnMissingField: throwOnMissingField,
                    rootId: id,
                    selectionSet: field.selectionSet,
                    variables: variables,
                    dataIdFromObject: dataIdFromObject,
                });
                if (!itemDiffResult.mergeUp) {
                    itemDiffResult.missingSelectionSets.forEach(function (itemSelectionSet) { return missingSelectionSets.push(itemSelectionSet); });
                }
                else {
                    pushMissingField(field);
                }
                return itemDiffResult.result;
            });
            return;
        }
        if (isString(storeValue)) {
            var subObjDiffResult = diffSelectionSetAgainstStore({
                store: store,
                throwOnMissingField: throwOnMissingField,
                rootId: storeValue,
                selectionSet: field.selectionSet,
                variables: variables,
                dataIdFromObject: dataIdFromObject,
            });
            if (!subObjDiffResult.mergeUp) {
                subObjDiffResult.missingSelectionSets.forEach(function (subObjSelectionSet) { return missingSelectionSets.push(subObjSelectionSet); });
            }
            else {
                pushMissingField(field);
            }
            result[resultFieldKey] = subObjDiffResult.result;
            return;
        }
        throw new Error('Unexpected number value in the store where the query had a subselection.');
    });
    var mergeUp = false;
    if (missingFields.length) {
        if (dataIdFromObject) {
            var id = dataIdFromObject(storeObj);
            if (typeof id !== 'string' && rootId !== 'ROOT_QUERY') {
                throw new Error("Can't generate query to refetch object " + rootId + ", since it doesn't have a string id.");
            }
            var typeName = void 0;
            if (rootId === 'ROOT_QUERY') {
                typeName = 'Query';
            }
            else if (!storeObj.__typename) {
                throw new Error("Can't generate query to refetch object " + rootId + ", since __typename wasn't in the store.");
            }
            else {
                typeName = storeObj.__typename;
            }
            missingSelectionSets.push({
                id: rootId,
                typeName: typeName,
                selectionSet: {
                    kind: 'SelectionSet',
                    selections: missingFields,
                },
            });
        }
        else if (rootId === 'ROOT_QUERY') {
            var typeName = 'Query';
            missingSelectionSets.push({
                id: rootId,
                typeName: typeName,
                selectionSet: {
                    kind: 'SelectionSet',
                    selections: missingFields,
                },
            });
        }
        else {
            mergeUp = true;
            missingSelectionSets.push({
                id: 'CANNOT_REFETCH',
                typeName: 'CANNOT_REFETCH',
                selectionSet: {
                    kind: 'SelectionSet',
                    selections: missingFields,
                },
            });
        }
    }
    return {
        result: result,
        missingSelectionSets: missingSelectionSets,
        mergeUp: mergeUp,
    };
}
exports.diffSelectionSetAgainstStore = diffSelectionSetAgainstStore;
//# sourceMappingURL=diffAgainstStore.js.map