"use strict";
var QueryManager_1 = require('../src/QueryManager');
var store_1 = require('../src/store');
var extensions_1 = require('../src/data/extensions');
var chai_1 = require('chai');
var async_1 = require('async');
var mockNetworkInterface_1 = require('./mocks/mockNetworkInterface');
describe('QueryManager', function () {
    it('properly roundtrips through a Redux store', function (done) {
        var query = "\n      query people {\n        allPeople(first: 1) {\n          people {\n            name\n          }\n        }\n      }\n    ";
        var data = {
            allPeople: {
                people: [
                    {
                        name: 'Luke Skywalker',
                    },
                ],
            },
        };
        var networkInterface = mockNetworkInterface_1.default({
            request: { query: query },
            result: { data: data },
        });
        var queryManager = new QueryManager_1.QueryManager({
            networkInterface: networkInterface,
            store: store_1.createApolloStore(),
            reduxRootKey: 'apollo',
        });
        var handle = queryManager.watchQuery({
            query: query,
        });
        handle.subscribe({
            next: function (result) {
                chai_1.assert.deepEqual(result.data, data);
                done();
            },
        });
    });
    it('properly roundtrips through a Redux store with variables', function (done) {
        var query = "\n      query people($firstArg: Int) {\n        allPeople(first: $firstArg) {\n          people {\n            name\n          }\n        }\n      }\n    ";
        var variables = {
            firstArg: 1,
        };
        var data = {
            allPeople: {
                people: [
                    {
                        name: 'Luke Skywalker',
                    },
                ],
            },
        };
        var networkInterface = mockNetworkInterface_1.default({
            request: { query: query, variables: variables },
            result: { data: data },
        });
        var queryManager = new QueryManager_1.QueryManager({
            networkInterface: networkInterface,
            store: store_1.createApolloStore(),
            reduxRootKey: 'apollo',
        });
        var handle = queryManager.watchQuery({
            query: query,
            variables: variables,
        });
        handle.subscribe({
            next: function (result) {
                chai_1.assert.deepEqual(result.data, data);
                done();
            },
        });
    });
    it('handles GraphQL errors', function (done) {
        var query = "\n      query people {\n        allPeople(first: 1) {\n          people {\n            name\n          }\n        }\n      }\n    ";
        var networkInterface = mockNetworkInterface_1.default({
            request: { query: query },
            result: {
                errors: [
                    {
                        name: 'Name',
                        message: 'This is an error message.',
                    },
                ],
            },
        });
        var queryManager = new QueryManager_1.QueryManager({
            networkInterface: networkInterface,
            store: store_1.createApolloStore(),
            reduxRootKey: 'apollo',
        });
        var handle = queryManager.watchQuery({
            query: query,
        });
        handle.subscribe({
            next: function (result) {
                chai_1.assert.equal(result.errors[0].message, 'This is an error message.');
                done();
            },
        });
    });
    it('handles network errors', function (done) {
        var query = "\n      query people {\n        allPeople(first: 1) {\n          people {\n            name\n          }\n        }\n      }\n    ";
        var networkInterface = mockNetworkInterface_1.default({
            request: { query: query },
            error: new Error('Network error'),
        });
        var queryManager = new QueryManager_1.QueryManager({
            networkInterface: networkInterface,
            store: store_1.createApolloStore(),
            reduxRootKey: 'apollo',
        });
        var handle = queryManager.watchQuery({
            query: query,
        });
        handle.subscribe({
            next: function (result) {
                done(new Error('Should not deliver result'));
            },
            error: function (error) {
                chai_1.assert.equal(error.message, 'Network error');
                done();
            },
        });
    });
    it('handles an unsubscribe action that happens before data returns', function (done) {
        var query = "\n      query people {\n        allPeople(first: 1) {\n          people {\n            name\n          }\n        }\n      }\n    ";
        var networkInterface = mockNetworkInterface_1.default({
            request: { query: query },
            delay: 1000,
        });
        var queryManager = new QueryManager_1.QueryManager({
            networkInterface: networkInterface,
            store: store_1.createApolloStore(),
            reduxRootKey: 'apollo',
        });
        var handle = queryManager.watchQuery({
            query: query,
        });
        var subscription = handle.subscribe({
            next: function (result) {
                done(new Error('Should not deliver result'));
            },
            error: function (error) {
                done(new Error('Should not deliver result'));
            },
        });
        chai_1.assert.doesNotThrow(subscription.unsubscribe);
        done();
    });
    it('allows you to refetch queries', function (done) {
        var query = "\n      {\n        people_one(id: 1) {\n          name\n        }\n      }\n    ";
        var data1 = {
            people_one: {
                name: 'Luke Skywalker',
            },
        };
        var data2 = {
            people_one: {
                name: 'Luke Skywalker has a new name',
            },
        };
        var networkInterface = mockNetworkInterface_1.default({
            request: { query: query },
            result: { data: data1 },
        }, {
            request: { query: query },
            result: { data: data2 },
        });
        var queryManager = new QueryManager_1.QueryManager({
            networkInterface: networkInterface,
            store: store_1.createApolloStore(),
            reduxRootKey: 'apollo',
        });
        var handleCount = 0;
        var handle = queryManager.watchQuery({
            query: query,
        });
        var subscription = handle.subscribe({
            next: function (result) {
                handleCount++;
                if (handleCount === 1) {
                    chai_1.assert.deepEqual(result.data, data1);
                    subscription.refetch();
                }
                else if (handleCount === 2) {
                    chai_1.assert.deepEqual(result.data, data2);
                    done();
                }
            },
        });
    });
    it('doesn\'t explode if you refetch before first fetch is done with query diffing', function (done) {
        var primeQuery = "\n      {\n        people_one(id: 1) {\n          name\n        }\n      }\n    ";
        var complexQuery = "\n      {\n        luke: people_one(id: 1) {\n          name\n        }\n        vader: people_one(id: 4) {\n          name\n        }\n      }\n    ";
        var diffedQuery = "\n      {\n        vader: people_one(id: 4) {\n          name\n        }\n      }\n    ";
        var data1 = {
            people_one: {
                name: 'Luke Skywalker',
            },
        };
        var data2 = {
            vader: {
                name: 'Darth Vader',
            },
        };
        var dataRefetch = {
            luke: {
                name: 'Luke has a new name',
            },
            vader: {
                name: 'Vader has a new name',
            },
        };
        var networkInterface = mockNetworkInterface_1.default({
            request: { query: primeQuery },
            result: { data: data1 },
        }, {
            request: { query: diffedQuery },
            result: { data: data2 },
            delay: 5,
        }, {
            request: { query: complexQuery },
            result: { data: dataRefetch },
            delay: 10,
        });
        var queryManager = new QueryManager_1.QueryManager({
            networkInterface: networkInterface,
            store: store_1.createApolloStore(),
            reduxRootKey: 'apollo',
        });
        queryManager.query({
            query: primeQuery,
        }).then(function () {
            var handleCount = 0;
            var handle = queryManager.watchQuery({
                query: complexQuery,
            });
            var subscription = handle.subscribe({
                next: function (result) {
                    handleCount++;
                    if (handleCount === 1) {
                        chai_1.assert.deepEqual(result.data, dataRefetch);
                        subscription.unsubscribe();
                        done();
                    }
                },
                error: function (error) {
                    done(error);
                },
            });
            subscription.refetch();
        });
    });
    it('runs a mutation', function () {
        var mutation = "\n      mutation makeListPrivate {\n        makeListPrivate(id: \"5\")\n      }\n    ";
        var data = {
            makeListPrivate: true,
        };
        var networkInterface = mockNetworkInterface_1.default({
            request: { query: mutation },
            result: { data: data },
        });
        var queryManager = new QueryManager_1.QueryManager({
            networkInterface: networkInterface,
            store: store_1.createApolloStore(),
            reduxRootKey: 'apollo',
        });
        return queryManager.mutate({
            mutation: mutation,
        }).then(function (result) {
            chai_1.assert.deepEqual(result.data, data);
        });
    });
    it('runs a mutation with variables', function () {
        var mutation = "\n      mutation makeListPrivate($listId: ID!) {\n        makeListPrivate(id: $listId)\n      }\n    ";
        var variables = {
            listId: '1',
        };
        var data = {
            makeListPrivate: true,
        };
        var networkInterface = mockNetworkInterface_1.default({
            request: { query: mutation, variables: variables },
            result: { data: data },
        });
        var queryManager = new QueryManager_1.QueryManager({
            networkInterface: networkInterface,
            store: store_1.createApolloStore(),
            reduxRootKey: 'apollo',
        });
        return queryManager.mutate({
            mutation: mutation,
            variables: variables,
        }).then(function (result) {
            chai_1.assert.deepEqual(result.data, data);
        });
    });
    it('runs a mutation and puts the result in the store', function () {
        var mutation = "\n      mutation makeListPrivate {\n        makeListPrivate(id: \"5\") {\n          id,\n          isPrivate,\n        }\n      }\n    ";
        var data = {
            makeListPrivate: {
                id: '5',
                isPrivate: true,
            },
        };
        var networkInterface = mockNetworkInterface_1.default({
            request: { query: mutation },
            result: { data: data },
        });
        var store = store_1.createApolloStore({
            config: { dataIdFromObject: extensions_1.getIdField },
        });
        var queryManager = new QueryManager_1.QueryManager({
            networkInterface: networkInterface,
            store: store,
            reduxRootKey: 'apollo',
        });
        return queryManager.mutate({
            mutation: mutation,
        }).then(function (result) {
            chai_1.assert.deepEqual(result.data, data);
            chai_1.assert.deepEqual(store.getState()['apollo'].data['5'], { id: '5', isPrivate: true });
        });
    });
    it('runs a mutation and puts the result in the store', function () {
        var mutation = "\n      mutation makeListPrivate {\n        makeListPrivate(id: \"5\") {\n          id,\n          isPrivate,\n        }\n      }\n    ";
        var data = {
            makeListPrivate: {
                id: '5',
                isPrivate: true,
            },
        };
        var networkInterface = mockNetworkInterface_1.default({
            request: { query: mutation },
            result: { data: data },
        });
        var store = store_1.createApolloStore({
            config: { dataIdFromObject: extensions_1.getIdField },
        });
        var queryManager = new QueryManager_1.QueryManager({
            networkInterface: networkInterface,
            store: store,
            reduxRootKey: 'apollo',
        });
        return queryManager.mutate({
            mutation: mutation,
        }).then(function (result) {
            chai_1.assert.deepEqual(result.data, data);
            chai_1.assert.deepEqual(store.getState()['apollo'].data['5'], { id: '5', isPrivate: true });
        });
    });
    it('runs a mutation and puts the result in the store with root key', function () {
        var mutation = "\n      mutation makeListPrivate {\n        makeListPrivate(id: \"5\") {\n          id,\n          isPrivate,\n        }\n      }\n    ";
        var data = {
            makeListPrivate: {
                id: '5',
                isPrivate: true,
            },
        };
        var networkInterface = mockNetworkInterface_1.default({
            request: { query: mutation },
            result: { data: data },
        });
        var reduxRootKey = 'test';
        var store = store_1.createApolloStore({
            reduxRootKey: reduxRootKey,
            config: { dataIdFromObject: extensions_1.getIdField },
        });
        var queryManager = new QueryManager_1.QueryManager({
            networkInterface: networkInterface,
            store: store,
            reduxRootKey: reduxRootKey,
        });
        return queryManager.mutate({
            mutation: mutation,
        }).then(function (result) {
            chai_1.assert.deepEqual(result.data, data);
            chai_1.assert.deepEqual(store.getState()[reduxRootKey].data['5'], { id: '5', isPrivate: true });
        });
    });
    it('diffs queries', function (done) {
        testDiffing([
            {
                query: "\n          {\n            people_one(id: \"1\") {\n              __typename,\n              id,\n              name\n            }\n          }\n        ",
                diffedQuery: "\n          {\n            people_one(id: \"1\") {\n              __typename,\n              id,\n              name\n            }\n          }\n        ",
                diffedQueryResponse: {
                    people_one: {
                        __typename: 'Person',
                        id: 'lukeId',
                        name: 'Luke Skywalker',
                    },
                },
                fullResponse: {
                    people_one: {
                        __typename: 'Person',
                        id: 'lukeId',
                        name: 'Luke Skywalker',
                    },
                },
                variables: {},
            },
            {
                query: "\n          {\n            people_one(id: \"1\") {\n              name\n              age\n            }\n          }\n        ",
                diffedQuery: "\n          {\n            __node_0: node(id: \"lukeId\") {\n              id\n              ... on Person {\n                age\n              }\n            }\n          }\n        ",
                diffedQueryResponse: {
                    __node_0: {
                        id: 'lukeId',
                        age: 45,
                    },
                },
                fullResponse: {
                    people_one: {
                        name: 'Luke Skywalker',
                        age: 45,
                    },
                },
                variables: {},
            },
            {
                query: "\n          {\n            people_one(id: \"1\") {\n              id\n              name\n              age\n            }\n          }\n        ",
                diffedQuery: null,
                diffedQueryResponse: null,
                fullResponse: {
                    people_one: {
                        id: 'lukeId',
                        name: 'Luke Skywalker',
                        age: 45,
                    },
                },
                variables: {},
            },
        ], {
            dataIdFromObject: extensions_1.getIdField,
        }, done);
    });
    it('diffs queries, preserving variable declarations', function (done) {
        testDiffing([
            {
                query: "\n          {\n            people_one(id: \"1\") {\n              __typename,\n              id,\n              name\n            }\n          }\n        ",
                diffedQuery: "\n          {\n            people_one(id: \"1\") {\n              __typename,\n              id,\n              name\n            }\n          }\n        ",
                diffedQueryResponse: {
                    people_one: {
                        __typename: 'Person',
                        id: '1',
                        name: 'Luke Skywalker',
                    },
                },
                fullResponse: {
                    people_one: {
                        __typename: 'Person',
                        id: '1',
                        name: 'Luke Skywalker',
                    },
                },
                variables: {},
            },
            {
                query: "\n          query getSeveralPeople($lukeId: String!, $vaderId: String!) {\n            luke: people_one(id: $lukeId) {\n              __typename\n              id\n              name\n            }\n            vader: people_one(id: $vaderId) {\n              __typename\n              id\n              name\n            }\n          }\n        ",
                diffedQuery: "\n          query getSeveralPeople($lukeId: String!, $vaderId: String!) {\n            vader: people_one(id: $vaderId) {\n              __typename\n              id\n              name\n            }\n          }\n        ",
                diffedQueryResponse: {
                    vader: {
                        __typename: 'Person',
                        id: '4',
                        name: 'Darth Vader',
                    },
                },
                fullResponse: {
                    luke: {
                        __typename: 'Person',
                        id: '1',
                        name: 'Luke Skywalker',
                    },
                    vader: {
                        __typename: 'Person',
                        id: '4',
                        name: 'Darth Vader',
                    },
                },
                variables: {
                    lukeId: '1',
                    vaderId: '4',
                },
            },
        ], {}, done);
    });
    it("doesn't return data while query is loading", function (done) {
        var query1 = "\n      {\n        people_one(id: 1) {\n          name\n        }\n      }\n    ";
        var data1 = {
            people_one: {
                name: 'Luke Skywalker',
            },
        };
        var query2 = "\n      {\n        people_one(id: 5) {\n          name\n        }\n      }\n    ";
        var data2 = {
            people_one: {
                name: 'Darth Vader',
            },
        };
        var networkInterface = mockNetworkInterface_1.default({
            request: { query: query1 },
            result: { data: data1 },
            delay: 10,
        }, {
            request: { query: query2 },
            result: { data: data2 },
        });
        var queryManager = new QueryManager_1.QueryManager({
            networkInterface: networkInterface,
            store: store_1.createApolloStore(),
            reduxRootKey: 'apollo',
        });
        var handle1 = queryManager.watchQuery({
            query: query1,
        });
        var handle2 = queryManager.watchQuery({
            query: query2,
        });
        var handle1Count = 0;
        var handle2Count = 0;
        handle1.subscribe({
            next: function (result) {
                handle1Count++;
                checkDone();
            },
        });
        handle2.subscribe({
            next: function (result) {
                handle2Count++;
                checkDone();
            },
        });
        function checkDone() {
            if (handle1Count === 1 && handle2Count === 2) {
                done();
            }
            if (handle1Count > 1) {
                chai_1.assert.fail();
            }
        }
    });
    it("updates result of previous query if the result of a new query overlaps", function (done) {
        var query1 = "\n      {\n        people_one(id: 1) {\n          name\n          age\n        }\n      }\n    ";
        var data1 = {
            people_one: {
                name: 'Luke Skywalker',
                age: 50,
            },
        };
        var query2 = "\n      {\n        people_one(id: 1) {\n          name\n          username\n        }\n      }\n    ";
        var data2 = {
            people_one: {
                name: 'Luke Skywalker has a new name',
                username: 'luke',
            },
        };
        var networkInterface = mockNetworkInterface_1.default({
            request: { query: query1 },
            result: { data: data1 },
        }, {
            request: { query: query2 },
            result: { data: data2 },
            delay: 10,
        });
        var queryManager = new QueryManager_1.QueryManager({
            networkInterface: networkInterface,
            store: store_1.createApolloStore(),
            reduxRootKey: 'apollo',
        });
        var handle1Count = 0;
        var handle1 = queryManager.watchQuery({
            query: query1,
        });
        handle1.subscribe({
            next: function (result) {
                handle1Count++;
                if (handle1Count === 1) {
                    chai_1.assert.deepEqual(result.data, data1);
                    queryManager.query({
                        query: query2,
                    });
                }
                else if (handle1Count === 3 &&
                    result.data['people_one'].name === 'Luke Skywalker has a new name') {
                    chai_1.assert.deepEqual(result.data, {
                        people_one: {
                            name: 'Luke Skywalker has a new name',
                            age: 50,
                        },
                    });
                    done();
                }
            },
        });
    });
});
function testDiffing(queryArray, config, done) {
    var mockedResponses = queryArray.map(function (_a) {
        var diffedQuery = _a.diffedQuery, diffedQueryResponse = _a.diffedQueryResponse, _b = _a.variables, variables = _b === void 0 ? {} : _b;
        return {
            request: { query: diffedQuery, variables: variables },
            result: { data: diffedQueryResponse },
        };
    });
    var networkInterface = mockNetworkInterface_1.default.apply(void 0, mockedResponses);
    var queryManager = new QueryManager_1.QueryManager({
        networkInterface: networkInterface,
        store: store_1.createApolloStore({
            config: { dataIdFromObject: extensions_1.getIdField },
        }),
        reduxRootKey: 'apollo',
        dataIdFromObject: config.dataIdFromObject,
    });
    var steps = queryArray.map(function (_a) {
        var query = _a.query, fullResponse = _a.fullResponse, variables = _a.variables;
        return function (cb) {
            queryManager.query({
                query: query,
                variables: variables,
            }).then(function (result) {
                chai_1.assert.deepEqual(result.data, fullResponse);
                cb();
            });
        };
    });
    async_1.series(steps, function (err, res) {
        if (err) {
            throw err;
        }
        done();
    });
}
//# sourceMappingURL=QueryManager.js.map